import numpy as np
from scipy.spatial import cKDTree
from .utils import cache

# https://en.wikipedia.org/wiki/ANSI_escape_code#SGR_(Select_Graphic_Rendition)_parameters
# section: CSI sequences
CUU_CODE = "\033[%dA"
CUD_CODE = "\033[%dB"

# https://en.wikipedia.org/wiki/ANSI_escape_code#SGR_(Select_Graphic_Rendition)_parameters
# section: SGR
SGR_8BIT_FG_COLOR_CODE = "\033[38;5;%dm"
SGR_8BIT_BG_COLOR_CODE = "\033[48;5;%dm"


SGR_8BIT_FG_COLOR_GRAY = "\033[38;5;%dm" % 243
SGR_8BIT_FG_COLOR_BLACK = "\033[38;5;%dm" % 0
SGR_8BIT_FG_COLOR_WHITE = "\033[38;5;%dm" % 15

SGR_8BIT_BG_COLOR_BLACK = "\033[48;5;%dm" % 0
SGR_8BIT_BG_COLOR_WHITE = "\033[48;5;%dm" % 15

STANDARD_8BIT_RGB = np.array(
    [
        [0x00, 0x00, 0x00],
        [0x80, 0x00, 0x00],
        [0x00, 0x80, 0x00],
        [0x80, 0x80, 0x00],
        [0x00, 0x00, 0x80],
        [0x80, 0x00, 0x80],
        [0x00, 0x80, 0x80],
        [0xC0, 0xC0, 0xC0],
        [0x80, 0x80, 0x80],
        [0xFF, 0x00, 0x00],
        [0x00, 0xFF, 0x00],
        [0xFF, 0xFF, 0x00],
        [0x00, 0x00, 0xFF],
        [0xFF, 0x00, 0xFF],
        [0x00, 0xFF, 0xFF],
        [0xFF, 0xFF, 0xFF],
        [0x00, 0x00, 0x00],
        [0x00, 0x00, 0x5F],
        [0x00, 0x00, 0x87],
        [0x00, 0x00, 0xAF],
        [0x00, 0x00, 0xD7],
        [0x00, 0x00, 0xFF],
        [0x00, 0x5F, 0x00],
        [0x00, 0x5F, 0x5F],
        [0x00, 0x5F, 0x87],
        [0x00, 0x5F, 0xAF],
        [0x00, 0x5F, 0xD7],
        [0x00, 0x5F, 0xFF],
        [0x00, 0x87, 0x00],
        [0x00, 0x87, 0x5F],
        [0x00, 0x87, 0x87],
        [0x00, 0x87, 0xAF],
        [0x00, 0x87, 0xD7],
        [0x00, 0x87, 0xFF],
        [0x00, 0xAF, 0x00],
        [0x00, 0xAF, 0x5F],
        [0x00, 0xAF, 0x87],
        [0x00, 0xAF, 0xAF],
        [0x00, 0xAF, 0xD7],
        [0x00, 0xAF, 0xFF],
        [0x00, 0xD7, 0x00],
        [0x00, 0xD7, 0x5F],
        [0x00, 0xD7, 0x87],
        [0x00, 0xD7, 0xAF],
        [0x00, 0xD7, 0xD7],
        [0x00, 0xD7, 0xFF],
        [0x00, 0xFF, 0x00],
        [0x00, 0xFF, 0x5F],
        [0x00, 0xFF, 0x87],
        [0x00, 0xFF, 0xAF],
        [0x00, 0xFF, 0xD7],
        [0x00, 0xFF, 0xFF],
        [0x5F, 0x00, 0x00],
        [0x5F, 0x00, 0x5F],
        [0x5F, 0x00, 0x87],
        [0x5F, 0x00, 0xAF],
        [0x5F, 0x00, 0xD7],
        [0x5F, 0x00, 0xFF],
        [0x5F, 0x5F, 0x00],
        [0x5F, 0x5F, 0x5F],
        [0x5F, 0x5F, 0x87],
        [0x5F, 0x5F, 0xAF],
        [0x5F, 0x5F, 0xD7],
        [0x5F, 0x5F, 0xFF],
        [0x5F, 0x87, 0x00],
        [0x5F, 0x87, 0x5F],
        [0x5F, 0x87, 0x87],
        [0x5F, 0x87, 0xAF],
        [0x5F, 0x87, 0xD7],
        [0x5F, 0x87, 0xFF],
        [0x5F, 0xAF, 0x00],
        [0x5F, 0xAF, 0x5F],
        [0x5F, 0xAF, 0x87],
        [0x5F, 0xAF, 0xAF],
        [0x5F, 0xAF, 0xD7],
        [0x5F, 0xAF, 0xFF],
        [0x5F, 0xD7, 0x00],
        [0x5F, 0xD7, 0x5F],
        [0x5F, 0xD7, 0x87],
        [0x5F, 0xD7, 0xAF],
        [0x5F, 0xD7, 0xD7],
        [0x5F, 0xD7, 0xFF],
        [0x5F, 0xFF, 0x00],
        [0x5F, 0xFF, 0x5F],
        [0x5F, 0xFF, 0x87],
        [0x5F, 0xFF, 0xAF],
        [0x5F, 0xFF, 0xD7],
        [0x5F, 0xFF, 0xFF],
        [0x87, 0x00, 0x00],
        [0x87, 0x00, 0x5F],
        [0x87, 0x00, 0x87],
        [0x87, 0x00, 0xAF],
        [0x87, 0x00, 0xD7],
        [0x87, 0x00, 0xFF],
        [0x87, 0x5F, 0x00],
        [0x87, 0x5F, 0x5F],
        [0x87, 0x5F, 0x87],
        [0x87, 0x5F, 0xAF],
        [0x87, 0x5F, 0xD7],
        [0x87, 0x5F, 0xFF],
        [0x87, 0x87, 0x00],
        [0x87, 0x87, 0x5F],
        [0x87, 0x87, 0x87],
        [0x87, 0x87, 0xAF],
        [0x87, 0x87, 0xD7],
        [0x87, 0x87, 0xFF],
        [0x87, 0xAF, 0x00],
        [0x87, 0xAF, 0x5F],
        [0x87, 0xAF, 0x87],
        [0x87, 0xAF, 0xAF],
        [0x87, 0xAF, 0xD7],
        [0x87, 0xAF, 0xFF],
        [0x87, 0xD7, 0x00],
        [0x87, 0xD7, 0x5F],
        [0x87, 0xD7, 0x87],
        [0x87, 0xD7, 0xAF],
        [0x87, 0xD7, 0xD7],
        [0x87, 0xD7, 0xFF],
        [0x87, 0xFF, 0x00],
        [0x87, 0xFF, 0x5F],
        [0x87, 0xFF, 0x87],
        [0x87, 0xFF, 0xAF],
        [0x87, 0xFF, 0xD7],
        [0x87, 0xFF, 0xFF],
        [0xAF, 0x00, 0x00],
        [0xAF, 0x00, 0x5F],
        [0xAF, 0x00, 0x87],
        [0xAF, 0x00, 0xAF],
        [0xAF, 0x00, 0xD7],
        [0xAF, 0x00, 0xFF],
        [0xAF, 0x5F, 0x00],
        [0xAF, 0x5F, 0x5F],
        [0xAF, 0x5F, 0x87],
        [0xAF, 0x5F, 0xAF],
        [0xAF, 0x5F, 0xD7],
        [0xAF, 0x5F, 0xFF],
        [0xAF, 0x87, 0x00],
        [0xAF, 0x87, 0x5F],
        [0xAF, 0x87, 0x87],
        [0xAF, 0x87, 0xAF],
        [0xAF, 0x87, 0xD7],
        [0xAF, 0x87, 0xFF],
        [0xAF, 0xAF, 0x00],
        [0xAF, 0xAF, 0x5F],
        [0xAF, 0xAF, 0x87],
        [0xAF, 0xAF, 0xAF],
        [0xAF, 0xAF, 0xD7],
        [0xAF, 0xAF, 0xFF],
        [0xAF, 0xD7, 0x00],
        [0xAF, 0xD7, 0x5F],
        [0xAF, 0xD7, 0x87],
        [0xAF, 0xD7, 0xAF],
        [0xAF, 0xD7, 0xD7],
        [0xAF, 0xD7, 0xFF],
        [0xAF, 0xFF, 0x00],
        [0xAF, 0xFF, 0x5F],
        [0xAF, 0xFF, 0x87],
        [0xAF, 0xFF, 0xAF],
        [0xAF, 0xFF, 0xD7],
        [0xAF, 0xFF, 0xFF],
        [0xD7, 0x00, 0x00],
        [0xD7, 0x00, 0x5F],
        [0xD7, 0x00, 0x87],
        [0xD7, 0x00, 0xAF],
        [0xD7, 0x00, 0xD7],
        [0xD7, 0x00, 0xFF],
        [0xD7, 0x5F, 0x00],
        [0xD7, 0x5F, 0x5F],
        [0xD7, 0x5F, 0x87],
        [0xD7, 0x5F, 0xAF],
        [0xD7, 0x5F, 0xD7],
        [0xD7, 0x5F, 0xFF],
        [0xD7, 0x87, 0x00],
        [0xD7, 0x87, 0x5F],
        [0xD7, 0x87, 0x87],
        [0xD7, 0x87, 0xAF],
        [0xD7, 0x87, 0xD7],
        [0xD7, 0x87, 0xFF],
        [0xD7, 0xAF, 0x00],
        [0xD7, 0xAF, 0x5F],
        [0xD7, 0xAF, 0x87],
        [0xD7, 0xAF, 0xAF],
        [0xD7, 0xAF, 0xD7],
        [0xD7, 0xAF, 0xFF],
        [0xD7, 0xD7, 0x00],
        [0xD7, 0xD7, 0x5F],
        [0xD7, 0xD7, 0x87],
        [0xD7, 0xD7, 0xAF],
        [0xD7, 0xD7, 0xD7],
        [0xD7, 0xD7, 0xFF],
        [0xD7, 0xFF, 0x00],
        [0xD7, 0xFF, 0x5F],
        [0xD7, 0xFF, 0x87],
        [0xD7, 0xFF, 0xAF],
        [0xD7, 0xFF, 0xD7],
        [0xD7, 0xFF, 0xFF],
        [0xFF, 0x00, 0x00],
        [0xFF, 0x00, 0x5F],
        [0xFF, 0x00, 0x87],
        [0xFF, 0x00, 0xAF],
        [0xFF, 0x00, 0xD7],
        [0xFF, 0x00, 0xFF],
        [0xFF, 0x5F, 0x00],
        [0xFF, 0x5F, 0x5F],
        [0xFF, 0x5F, 0x87],
        [0xFF, 0x5F, 0xAF],
        [0xFF, 0x5F, 0xD7],
        [0xFF, 0x5F, 0xFF],
        [0xFF, 0x87, 0x00],
        [0xFF, 0x87, 0x5F],
        [0xFF, 0x87, 0x87],
        [0xFF, 0x87, 0xAF],
        [0xFF, 0x87, 0xD7],
        [0xFF, 0x87, 0xFF],
        [0xFF, 0xAF, 0x00],
        [0xFF, 0xAF, 0x5F],
        [0xFF, 0xAF, 0x87],
        [0xFF, 0xAF, 0xAF],
        [0xFF, 0xAF, 0xD7],
        [0xFF, 0xAF, 0xFF],
        [0xFF, 0xD7, 0x00],
        [0xFF, 0xD7, 0x5F],
        [0xFF, 0xD7, 0x87],
        [0xFF, 0xD7, 0xAF],
        [0xFF, 0xD7, 0xD7],
        [0xFF, 0xD7, 0xFF],
        [0xFF, 0xFF, 0x00],
        [0xFF, 0xFF, 0x5F],
        [0xFF, 0xFF, 0x87],
        [0xFF, 0xFF, 0xAF],
        [0xFF, 0xFF, 0xD7],
        [0xFF, 0xFF, 0xFF],
        [0x08, 0x08, 0x08],
        [0x12, 0x12, 0x12],
        [0x1C, 0x1C, 0x1C],
        [0x26, 0x26, 0x26],
        [0x30, 0x30, 0x30],
        [0x3A, 0x3A, 0x3A],
        [0x44, 0x44, 0x44],
        [0x4E, 0x4E, 0x4E],
        [0x58, 0x58, 0x58],
        [0x62, 0x62, 0x62],
        [0x6C, 0x6C, 0x6C],
        [0x76, 0x76, 0x76],
        [0x80, 0x80, 0x80],
        [0x8A, 0x8A, 0x8A],
        [0x94, 0x94, 0x94],
        [0x9E, 0x9E, 0x9E],
        [0xA8, 0xA8, 0xA8],
        [0xB2, 0xB2, 0xB2],
        [0xBC, 0xBC, 0xBC],
        [0xC6, 0xC6, 0xC6],
        [0xD0, 0xD0, 0xD0],
        [0xDA, 0xDA, 0xDA],
        [0xE4, 0xE4, 0xE4],
        [0xEE, 0xEE, 0xEE],
    ]
)

STANDARD_8BIT_RGB_KD_TREE = cKDTree(STANDARD_8BIT_RGB)


@cache(key=lambda args: args[0])
def fitting_sgr(rgb):
    _, index = STANDARD_8BIT_RGB_KD_TREE.query(rgb)
    return index
